import typing as tp
import os
import pickle

from satella.coding import Monitor

from smok.extras.sensors_database import BaseSensorDatabase
from smok.sensor import Sensor


def from_dict(device: 'SMOKDevice', data: tuple):
    return Sensor(device, *data)


def to_dict(s: Sensor) -> tuple:
    return s.fqts, s.path, s.type_name


class PicklingSensorDatabase(BaseSensorDatabase, Monitor):
    def __init__(self, path: str):
        super().__init__()
        Monitor.__init__(self)
        self.sensor_cache = {}
        if os.path.exists(path):
            try:
                with open(path, 'rb') as f_in:
                    self.data = pickle.load(f_in)
            except pickle.UnpicklingError:
                self.data = {}

    @Monitor.synchronized
    def get_sensor(self, fqts: str) -> Sensor:
        if fqts not in self.sensor_cache:
            sensor = from_dict(self.device, self.data[fqts])
            self.sensor_cache[fqts] = sensor
        return self.sensor_cache[fqts]

    @Monitor.synchronized
    def on_sensors_sync(self, sensors: tp.List[Sensor]):
        self.data = {}
        for sensor in sensors:
            self.data[sensor.fqts] = to_dict(sensor)
        self.save()

    def get_all_sensors(self) -> tp.Iterator[Sensor]:
        for key in self.data.keys():
            yield self.get_sensor(key)

    def save(self):
        with open(self.path, 'wb') as f_out:
            pickle.dump(self.data, f_out, pickle.HIGHEST_PROTOCOL)




import collections
import pickle
import time
import typing as tp

from satella.coding import silence_excs
from satella.coding.structures import OmniHashableMixin, ReprableMixin

from smok.pathpoint.orders import Section, WriteOrder, AdviseLevel
from smok.pathpoint.typing import PathpointValueType


class Macro(OmniHashableMixin, ReprableMixin):
    _HASH_FIELDS_TO_USE = ('macro_id',)
    _REPR_FIELDS = ('macro_id', 'commands', 'occurrences_not_done')
    __slots__ = ('macro_id', 'commands', 'occurrences_not_done')

    @classmethod
    def from_json(cls, dct: dict) -> 'Macro':
        macro_id = dct['macro_id']
        commands = {}
        for command in dct['commands']:
            commands[command['path']] = command['value']
        occurrences = []
        for occ in dct['occurrences']:
            if not occ['completed']:
                occurrences.append(occ['timestamp'])

        return Macro(macro_id, commands, occurrences)

    def __init__(self, macro_id: str,
                 commands: tp.Dict[str, PathpointValueType],
                 occurrences_not_done: tp.List[float]):
        self.macro_id = macro_id
        self.commands = commands
        self.occurrences_not_done = collections.deque(sorted(occurrences_not_done))

    def __getstate__(self):
        return {'macro_id': self.macro_id,
                'commands': self.commands,
                'occurrences_not_done': list(self.occurrences_not_done)}

    def __setstate__(self, state):
        self.macro_id = state['macro_id']
        self.commands = state['commands']
        self.occurrences_not_done = collections.deque(state['occurrences_not_done'])

    def to_pickle(self) -> bytes:
        """
        :return: self, pickled
        """
        return pickle.dumps(self, pickle.HIGHEST_PROTOCOL)

    @classmethod
    def from_pickle(cls, y: bytes) -> 'Macro':
        return pickle.loads(y)

    def __bool__(self) -> bool:
        return bool(self.occurrences_not_done)

    @silence_excs(IndexError, returns=False)
    def should_execute(self) -> bool:
        return time.time() > self.occurrences_not_done[0]

    def execute(self, device: 'SMOKDevice') -> None:
        while self.should_execute():
            ts = self.occurrences_not_done.popleft()
            sec = Section([WriteOrder(pathpoint_name, pathpoint_value, AdviseLevel.FORCE)
                           for pathpoint_name, pathpoint_value in self.commands.items()])
            device.execute(sec)
            device.macros_database.notify_macro_executed(self.macro_id, ts)




import logging
import typing as tp
import weakref

from satella.coding import rethrow_as
from satella.coding.transforms import merge_series
from satella.coding.typing import Number
from satella.time import time_ms

from smok.exceptions import NotReadedError, OperationFailedError
from smok.pathpoint import AdviseLevel, PathpointValueType
from smok.pathpoint.orders import Section, ReadOrder
from smok.sensor import reparse_funs
from .write_logs import SensorWriteEvent
from smok.sensor.reparse import parse
from smok.sensor.types import SensorValueType, get_type, SVTOrExcept


logger = logging.getLogger(__name__)


def fqtsify(tag_name: tp.Union[str, tp.Set[str]]) -> str:
    """
    Standarize the name, for use in dictionaries and other places that access sensors by name

    :param tag_name: :term:`Tag name`, either a space-separated set of names or a set of names
        proper
    :return: :term:`FQTS`-ified name
    """
    if isinstance(tag_name, set):
        tag_name = list(tag_name)
    elif isinstance(tag_name, str):
        tag_name = tag_name.split(' ')
    tag_name.sort()
    return ' '.join(tag_name)


class Sensor:
    """
    A class representing a smok-client sensor defined on given device.

    Class is eq-able and hashable by fqts.

    .. warning:: Do not compare sensors across different devices!

    :ivar fqts: this sensor's :term:`FQTS` (str)
    :ivar path: composite pathpoint names, separated by a tilde (~) (str)
    :ivar type_name: name of the sensor type (str)
    :ivar type: object used for data conversion between pathpoints and sensor values
        (:class:`smok.sensor.types.BasicType`)
    """
    __slots__ = ('fqts', 'path', 'type_name', '_pathpoint_names', 'type', 'device',
                 'slave_pathpoints')

    def __eq__(self, other) -> bool:
        if not isinstance(other, Sensor):
            return False
        return self.fqts == other.fqts

    def __hash__(self) -> int:
        return hash(self.fqts)

    def __init__(self, device: 'SMOKDevice', fqts: str, path: str, type_name: str):
        self.device = weakref.proxy(device)
        self.fqts = fqtsify(fqts)
        self.path = path
        self.type_name = type_name
        self.slave_pathpoints = []
        self._pathpoint_names = []
        for pn in path.split('~'):
            self._pathpoint_names.append(pn)
            self.slave_pathpoints.append(self.device.get_pathpoint(pn))
        self.type = get_type(type_name)

    def get_archive(self,
                    starting_at: int,
                    stopping_at: tp.Optional[int] = None) -> tp.Iterator[
            tp.Tuple[int, SVTOrExcept]]:
        """
        Get archive readings.

        This will be readed entirely from the device's data, the server will not be queried

        :param starting_at: timestamp of start, in milliseconds
        :param stopping_at: timestamp of end, in milliseconds, or None for the end of the park
        :return: an iterator of tuple (timestamp in milliseconds,
                                       pathpoint value or exception instance)
        """
        archives = []
        for slave in self.slave_pathpoints:
            archives.append(slave.get_archive(starting_at, stopping_at))
        merged = merge_series(*archives)
        for ts, *values in merged:
            for v in values:
                if isinstance(v, OperationFailedError):
                    yield ts, v
                    break
            else:
                yield ts, self.type.pathpoint_to_sensor(*values)

    def log_write(self, who: str, reason: str, value: str, hr_value: tp.Optional[str] = None,
                  hr_sensor: tp.Optional[str] = None, timestamp: tp.Optional[int] = None) -> None:
        """
        Log that a write has taken place in this sensor

        :param who: who has made this change
        :param reason: reason for this write
        :param value: a str-able form of value
        :param hr_value: a human-readable value. Defaults to value
        :param hr_sensor: a human-readable name for the sensor. Defaults to fqts
        :param timestamp: in milliseconds! Leave at default for current timestamp
        """
        if hr_value is None:
            hr_value = value
        if hr_sensor is None:
            hr_sensor = self.fqts
        if timestamp is None:
            timestamp = time_ms()
        perm = SensorWriteEvent(fqts=self.fqts, hr_sensor=hr_sensor, hr_value=hr_value, value=value,
                                reason=reason, who=who, timestamp=timestamp)
        self.device.log_sensor_write(perm)

    def write(self, value, advise: AdviseLevel = AdviseLevel.ADVISE) -> Section:
        """
        Write a particular value to the sensor.

        Take care for the value to match the type of the sensor

        :param value: value to write
        :param advise: advise level to use
        :raises TypeError: invalid type
        """
        values = self.type.sensor_to_pathpoint(value)
        section = Section()
        for value, pp in zip(values, self.slave_pathpoints):
            section += pp.write(value, advise)
        return section

    def get(self) -> tp.Tuple[Number, SensorValueType]:
        """
        Return the value of this sensor

        :raises NotReadedError: required pathpoints are not available
        :return: a tuple of (timestamp in milliseconds, sensor value)
        :raises OperationFailedError: one of pathpoint failed to provide a value
        """
        vals = [pp.get() for pp in self.slave_pathpoints]
        cur_ts = max(ts[0] for ts in vals)
        return cur_ts, self.type.pathpoint_to_sensor(*(val[1] for val in vals))

    def read(self, advise: AdviseLevel = AdviseLevel.ADVISE) -> Section:
        return Section([ReadOrder(y, advise) for y in self._pathpoint_names])

    @classmethod
    def from_json(cls, device: 'SMOKDevice', x: dict) -> 'Sensor':
        return Sensor(device, x['fqts'], x['path'], x['type'])



import typing as tp

from satella.coding import Monitor

from .base import BaseSensorDatabase
from ...sensor import Sensor


class InMemorySensorDatabase(BaseSensorDatabase, Monitor):
    @Monitor.synchronized
    def get_all_sensors(self) -> tp.Iterator[Sensor]:
        return list(self.sensors.values())

    @Monitor.synchronized
    def on_sensors_sync(self, sensors: tp.List[Sensor]):
        self.sensors = {sensor.fqts: sensor for sensor in sensors}

    @Monitor.synchronized
    def get_sensor(self, fqts: str) -> Sensor:
        return self.sensors[fqts]

    def __init__(self):
        super().__init__()
        Monitor.__init__(self)
        self.sensors = {}       # type: tp.Dict[str, Sensor]



import typing as tp
import json

import requests
from satella.files import read_in_file

from smok.basics import Environment
from smok.exceptions import ResponseError


class RequestsAPI:
    __slots__ = ('environment', 'base_url', 'cert')

    def __init__(self, device):
        self.environment = device.environment
        self.base_url = device.url
        if self.environment == Environment.STAGING:
            self.cert = read_in_file(device.cert[0], 'utf-8').replace('\n', '\t')
        else:
            self.cert = device.cert

    def request(self, request_type: str, url: str,
                direct_response: bool = False, **kwargs) -> tp.Union[dict, tp.Tuple[bytes, dict]]:
        """
        :param request_type: type of request, in lowercase
        :param url: URL to contact
        :param direct_response: if True then return a tuple of (
            response as is, headers), else it's JSON

        :raises ResponseError: something went wrong
        """
        op = getattr(requests, request_type)
        try:
            if self.environment == Environment.PRODUCTION:
                resp = op(self.base_url + url, cert=self.cert, **kwargs)
            else:
                resp = op(self.base_url + url, headers={
                    'X-SSL-Client-Certificate': self.cert
                }, **kwargs)
        except requests.RequestException as e:
            raise ResponseError(599, 'Requests error: %s' % (str(e), ))

        if resp is None and self.environment == Environment.LOCAL_DEVELOPMENT:
            # special case for CI
            if direct_response:
                return '', 200
            else:
                return {}

        if resp.status_code not in (200, 201):
            if direct_response:
                return resp.content, resp.headers
            try:
                error_json = resp.json()['status']
            except (json.decoder.JSONDecodeError, ValueError):
                try:
                    error_json = resp.text
                except AttributeError:
                    error_json = repr(resp.content)
            raise ResponseError(resp.status_code, error_json)
        if direct_response:
            return resp.content, resp.headers
        else:
            try:
                return resp.json()
            except json.decoder.JSONDecodeError:
                raise ResponseError(resp.status_code, resp.content)

    def get(self, url: str, **kwargs):
        return self.request('get', url, **kwargs)

    def post(self, url: str, **kwargs):
        return self.request('post', url, **kwargs)

    def put(self, url: str, **kwargs):
        return self.request('put', url, **kwargs)

    def patch(self, url: str, **kwargs):
        return self.request('patch', url, **kwargs)

    def delete(self, url: str, **kwargs):
        return self.request('delete', url, **kwargs)



import os
import pickle
import typing as tp
import weakref

from satella.coding import Monitor, silence_excs

from .base import BaseEventDatabase, BaseEventSynchronization
from ...predicate.event import Event


class InMemoryEventSynchronization(BaseEventSynchronization):
    __slots__ = ('event_db', 'events',)

    def __init__(self, event_db: 'InMemoryEventDatabase', events: tp.List[Event]):
        self.event_db = weakref.proxy(event_db)
        self.events = events

    def get_events(self) -> tp.List[Event]:
        return self.events

    def acknowledge(self, *uuids: str) -> None:
        for event, uuid in zip(self.events, uuids):
            if event.uuid is None:
                event.uuid = uuid

            if event.is_closed():
                with Monitor.acquire(self.event_db), silence_excs(ValueError):
                    self.events.remove(event)

    def negative_acknowledge(self) -> None:
        pass


class InMemoryEventDatabase(BaseEventDatabase, Monitor):

    @Monitor.synchronized
    def get_open_events(self) -> tp.Iterator[Event]:
        for event in self.events:
            if not event.is_closed():
                yield event

    @Monitor.synchronized
    def close_event(self, event: Event) -> None:
        self.events_to_sync.append(event)

    @Monitor.synchronized
    def get_events_to_sync(self) -> tp.Optional[BaseEventSynchronization]:
        if not self.events_to_sync:
            return None
        else:
            return InMemoryEventSynchronization(self, self.events_to_sync)

    @Monitor.synchronized
    def add_event(self, event: Event) -> None:
        self.events.append(event)
        self.events_to_sync.append(event)

    def __init__(self, path: str):
        self.events = []
        self.path = path
        self.internal_data = {}
        self.events_to_sync = []
        Monitor.__init__(self)
        if os.path.exists(self.path):
            with open(self.path, 'rb') as f_in:
                try:
                    self.internal_data = pickle.load(f_in)
                except pickle.PickleError:
                    pass

    def checkpoint(self) -> None:
        pass

    def set_cache(self, predicate_id: str, cache) -> None:
        self.internal_data[predicate_id] = cache
        self.sync()

    def get_cache(self, predicate_id: str) -> tp.Any:
        return self.internal_data[predicate_id]

    def sync(self):
        with open(self.path, 'wb') as f_out:
            pickle.dump(self.internal_data, f_out, pickle.HIGHEST_PROTOCOL)

    @silence_excs(KeyError)
    def on_predicate_deleted(self, predicate_id: str) -> None:
        del self.internal_data[predicate_id]
        self.sync()



from smok.exceptions import UnavailableError


class BAOB:
    """
    A Binary All-sized OBject.
    """

    __slots__ = ('device', 'key')

    def __init__(self, device: 'SMOKDevice', key: str):
        self.device = device
        self.key = key

    @property
    def version(self) -> int:
        """
        :return: version of this BAOB
        """
        try:
            return self.device.baob_database.get_baob_version(self.key)
        except KeyError:
            return -1

    @property
    def value(self) -> bytes:
        """
        This property is both gettable and settable.

        :return: value of this BAOB
        """
        return self.device.baob_database.get_baob_value(self.key)

    @value.setter
    def value(self, new_data: bytes) -> None:
        """
        :raises UnavailableError: client was launched in a mode with BAOBs disabled
        """
        if self.device.dont_do_baobs:
            raise UnavailableError('Client was launched as not supporting BAOBs!')
        self.device.baob_database.set_baob_value(self.key, new_data, self.version+1)
        self.device.getter.last_baob_synced = 0
        self.device.getter.data_to_update.notify()





import typing as tp

from satella.coding import Monitor, silence_excs

from .base import BaseBAOBDatabase


class InMemoryBAOBDatabase(BaseBAOBDatabase, Monitor):
    @Monitor.synchronized
    def set_baob_value(self, key: str, data: bytes, version: int):
        self.baobs_values[key] = data
        self.baobs_versions[key] = version

    @Monitor.synchronized
    def get_all_keys(self) -> tp.Iterator[str]:
        return list(self.baobs_values.keys())

    @Monitor.synchronized
    def get_baob_version(self, key: str) -> int:
        return self.baobs_versions[key]

    @Monitor.synchronized
    @silence_excs(KeyError)
    def delete_baob(self, key: str) -> None:
        del self.baobs_values[key]
        del self.baobs_versions[key]

    @Monitor.synchronized
    def get_baob_value(self, key: str) -> bytes:
        return self.baobs_values[key]

    def __init__(self):
        self.baobs_values = {}
        self.baobs_versions = {}
        Monitor.__init__(self)




import logging
import typing as tp

from satella.coding.transforms import merge_series
from satella.coding.typing import Number
from satella.exceptions import ImpossibleError

from . import AdviseLevel, Section, PathpointValueType, ValueOrExcept
from .pathpoint import Pathpoint
from ..basics import StorageLevel
from ..exceptions import OperationFailedError
from ..sensor.reparse import parse
from ..sensor.reparse_funs import ecre_eval


logger = logging.getLogger(__name__)


class ReparsePathpoint(Pathpoint):
    """
    A pathpoint whose value is derived from other pathpoints.

    It's first letter is 'r' then there goes the type letter and a Python expression
    with other pathpoints substituted by braces, so eg. rf{W1r4002}+2 would be
    pathpoint "W1r4002" plus two.

    Writing a reparse pathpoint always results in a zero operation, while reading
    it reads all constituent pathpoints.

    Good news is that you don't need to override it, nor provide your own implementations.
    """
    __slots__ = ('expr', 'slave_pathpoints')

    def __init__(self, device: tp.Optional['SMOKDevice'], name: str,
                 storage_level: StorageLevel = StorageLevel.TREND):
        super().__init__(device, name, storage_level)
        self.expr, pps = parse(name[2:])
        self.slave_pathpoints = []
        for pp in pps:
            self.slave_pathpoints.append(device.get_pathpoint(pp, storage_level))

    def on_new_storage_level(self, new_storage_level: StorageLevel) -> None:
        """
        Called when this pathpoint's storage level is updated.

        :param new_storage_level: new storage level for this pathpoint
        """
        for slave in self.slave_pathpoints:
            new_storage_level = slave.storage_level or new_storage_level
            if new_storage_level != slave.storage_level:
                slave.on_new_storage_level(new_storage_level)
        self.storage_level = new_storage_level

    def get_archive(self,
                    starting_at: int,
                    stopping_at: tp.Optional[int] = None) -> tp.Iterator[
            tp.Tuple[int, ValueOrExcept]]:
        """
        Get archive readings.

        This will be readed entirely from the device's data, the server will not be queried

        :param starting_at: timestamp of start, in milliseconds
        :param stopping_at: timestamp of end, in milliseconds, or None for the end of the park
        :return: an iterator of tuple (timestamp in milliseconds,
                                       sensor value or exception instance)
        """
        slave_archives = []
        for slave in self.slave_pathpoints:
            slave_archives.append(slave.get_archive(starting_at, stopping_at))
        merged_series = merge_series(*slave_archives)
        for ts, *values in merged_series:
            for v in values:
                if isinstance(v, OperationFailedError):
                    yield ts, v
                    break
            else:
                yield ts, ecre_eval(self.expr, args=values)

    def set_new_value(self, timestamp: Number, value: ValueOrExcept) -> None:
        """
        A no-op
        """

    def get(self) -> tp.Tuple[Number, PathpointValueType]:
        """
        Calculate the reparse pathpoint's value

        :return: a tuple of (timestamp in milliseconds, value)
        :raises OperationFailedError: when any constituent's pathpoint's read has failed
        """
        timestamps, values = [], []
        for slave in self.slave_pathpoints:
            ts, v = slave.get()
            timestamps.append(ts)
            values.append(v)
        return max(timestamps), ecre_eval(self.expr, args=values)

    def write(self, value, advise_level=AdviseLevel.ADVISE,
              stale_after=None) -> Section:
        """
        Writing a reparse pathpoint is always a zero operation
        """
        return Section()

    def read(self, advise_level=AdviseLevel.ADVISE):
        """
        Reading a reparse pathpoint results in all of it's constituents being read.
        """
        sec = Section()
        for slave in self.slave_pathpoints:
            sec += slave.read(advise_level)
        return sec

    def on_read(self, advise):
        """
        It is an ImpossibleError to call this
        """
        raise ImpossibleError('should never be called!')

    def on_write(self, value, advise):
        """
        It is an ImpossibleError to call this
        """
        raise ImpossibleError('should never be called!')
